@@ -29,6 +29,7 @@ var WritableStream = require('./mockWritableStream');
 var EventEmitter = require('./mockEventEmitter');
 var mime = require('mime');
 var http = require('./node/http');
+var utils = require('./utils');
 
 function createResponse(options) {
 
@@ -38,6 +39,9 @@ function createResponse(options) {
 
     var _endCalled = false;
     var _data = '';
+    var _buffer = Buffer.alloc(0);
+    var _chunks = [];
+    var _size = 0;
     var _encoding = options.encoding;
 
     var _redirectUrl = '';
@@ -51,16 +55,25 @@ function createResponse(options) {
         EventEmitter = options.eventEmitter;
     }
     var writableStream = new WritableStream();
-    var eventEmitter = new EventEmitter();
 
+    var mockRequest = options.req;
     // create mockResponse
 
-    var mockResponse = {};
+    var mockResponse = Object.create(EventEmitter.prototype);
+    EventEmitter.call(mockResponse);
 
     mockResponse._headers = {};
 
-    mockResponse.statusCode = 200;
     mockResponse.cookies = {};
+    mockResponse.finished = false;
+    mockResponse.writableEnded = false;
+    mockResponse.writableFinished = false;
+    mockResponse.headersSent = false;
+    mockResponse.statusCode = 200;
+    mockResponse.statusMessage = 'OK';
+
+    // http://expressjs.com/en/api.html#res.locals
+    mockResponse.locals = options.locals || {};
 
     mockResponse.cookie = function(name, value, opt) {
 
@@ -69,10 +82,15 @@ function createResponse(options) {
             options: opt
         };
 
+        return this;
     };
 
-    mockResponse.clearCookie = function(name) {
-        delete mockResponse.cookies[name];
+    mockResponse.clearCookie = function(name, opt) {
+        var opts = opt || {};
+        opts.expires = new Date(1);
+        opts.path = '/';
+
+        return this.cookie(name, '', opts);
     };
 
     mockResponse.status = function(code) {
@@ -91,28 +109,49 @@ function createResponse(options) {
      *  headers    - An object of properties that will be used for
      *               the HTTP headers.
      */
-    mockResponse.writeHead = function(statusCode, phrase, headers) {
+    mockResponse.writeHead = function(statusCode, statusMessage, headers) {
 
         if (_endCalled) {
-            throw 'The end() method has already been called.';
+            throw new Error('The end() method has already been called.');
+        }
+
+        if (mockResponse.headersSent) {
+            // Node docs: "This method must only be called once on a message"
+            // but it doesn't error if you do call it after first chunk of body is sent
+            // so we shouldn't throw here either (although it's a bug in the code).
+            // We return without updating since in real life it's just possible the double call didn't
+            // completely corrupt the response (for example not using chunked encoding due to HTTP/1.0 client)
+            // and in this case the client will see the _original_ headers.
+            return;
         }
 
         mockResponse.statusCode = statusCode;
 
-        // Note: Not sure if the headers given in this function
-        //       overwrite any headers specified earlier.
+        // resolve statusMessage and headers as optional
+        if (Object.prototype.toString.call(statusMessage) === '[object Object]') {
+            headers = statusMessage;
+            statusMessage = null;
+        }
+
+        if (statusMessage) {
+            mockResponse.statusMessage = statusMessage;
+        }
+
+        // The headers specified earlier (been set with `mockResponse.setHeader`)
+        // should not be overwritten but be merged with the headers
+        // passed into `mockResponse.writeHead`.
         if (headers) {
-            mockResponse._headers = headers;
-        } else {
-            mockResponse._headers = phrase;
+            Object.assign(mockResponse._headers, utils.convertKeysToLowerCase(headers));
         }
 
     };
 
     /**
-     *  The 'send' function from node's HTTP API that returns data
+     *  The 'send' function from restify's Response API that returns data
      *  to the client. Can be called multiple times.
      *
+     *  @see http://mcavage.me/node-restify/#response-api
+     *
      * @param data The data to return. Must be a string.
      */
     mockResponse.send = function(a, b, c) {
@@ -124,7 +163,7 @@ function createResponse(options) {
                 if (data.statusCode) {
                     mockResponse.statusCode = data.statusCode;
                 } else if (data.httpCode) {
-                    mockResponse.statusCode = data.statusCode;
+                    mockResponse.statusCode = data.httpCode;
                 }
 
                 if (data.body) {
@@ -140,19 +179,15 @@ function createResponse(options) {
         };
 
         switch (arguments.length) {
-
             case 1:
-
                 if (typeof a === 'number') {
                     mockResponse.statusCode = a;
                 } else {
                     _formatData(a);
                 }
-
                 break;
 
             case 2:
-
                 if (typeof a === 'number') {
                     _formatData(b);
                     mockResponse.statusCode = a;
@@ -164,26 +199,25 @@ function createResponse(options) {
                     _formatData(a);
                     _encoding = b;
                 }
-
                 break;
 
             case 3:
-
                 _formatData(a);
-                mockResponse._headers = b;
+                mockResponse._headers = utils.convertKeysToLowerCase(b);
                 mockResponse.statusCode = c;
                 console.warn('WARNING: Called send() with deprecated three parameters');
-
                 break;
 
             default:
                 break;
-
         }
 
-        mockResponse.emit('send');
-        mockResponse.emit('end');
+        mockResponse.headersSent = true;
 
+        mockResponse.emit('send');
+        mockResponse.end();
+
+        return mockResponse;
     };
 
     /**
@@ -228,24 +262,21 @@ function createResponse(options) {
     mockResponse.json = function(a, b) {
 
         mockResponse.setHeader('Content-Type', 'application/json');
-        if (a) {
-            if (typeof a === 'number') {
+        if (typeof a !== 'undefined') {
+            if (typeof a === 'number' && typeof b !== 'undefined') {
                 mockResponse.statusCode = a;
-            } else {
-                _data += JSON.stringify(a);
-            }
-        }
-        if (b) {
-            if (typeof b === 'number') {
+                mockResponse.write(JSON.stringify(b), 'utf8');
+            } else if(typeof b !== 'undefined' && typeof b === 'number') {
                 mockResponse.statusCode = b;
+                mockResponse.write(JSON.stringify(a), 'utf8');
             } else {
-                _data += JSON.stringify(b);
+                mockResponse.write(JSON.stringify(a), 'utf8');
             }
         }
-
         mockResponse.emit('send');
-        mockResponse.emit('end');
+        mockResponse.end();
 
+        return mockResponse;
     };
 
     /**
@@ -265,25 +296,21 @@ function createResponse(options) {
     mockResponse.jsonp = function(a, b) {
 
         mockResponse.setHeader('Content-Type', 'text/javascript');
-
-        if (a) {
-            if (typeof a === 'number') {
+        if (typeof a !== 'undefined') {
+            if (typeof a === 'number' && typeof b !== 'undefined') {
                 mockResponse.statusCode = a;
+                _data += JSON.stringify(b);
+            } else if(typeof b !== 'undefined' && typeof b === 'number') {
+                mockResponse.statusCode = b;
+                _data += JSON.stringify(a);
             } else {
                 _data += JSON.stringify(a);
             }
         }
-        if (b) {
-            if (typeof b === 'number') {
-                mockResponse.statusCode = b;
-            } else {
-                _data += JSON.stringify(b);
-            }
-        }
-
         mockResponse.emit('send');
-        mockResponse.emit('end');
+        mockResponse.end();
 
+        return mockResponse;
     };
 
     /**
@@ -302,10 +329,22 @@ function createResponse(options) {
      * @return {ServerResponse} for chaining
      * @api public
      */
-    mockResponse.contentType = mockResponse.type = function(type){
+    mockResponse.contentType = mockResponse.type = function(type) {
         return mockResponse.set('Content-Type', type.indexOf('/') >= 0 ? type : mime.lookup(type));
     };
 
+    /**
+     * Set 'Location' response header.
+     *
+     * @see http://expressjs.com/en/api.html#res.location
+     *
+     * @param  {String}         location The location to set in the header.
+     * @return {ServerResponse}          For chaining
+     */
+    mockResponse.location = function(location) {
+        return mockResponse.set('Location', location);
+    };
+
     /**
      * Function: write
      *
@@ -320,7 +359,14 @@ function createResponse(options) {
 
     mockResponse.write = function(data, encoding) {
 
-        _data += data;
+        mockResponse.headersSent = true;
+
+        if (data instanceof Buffer) {
+            _chunks.push(data);
+            _size += data.length;
+        } else {
+            _data += data;
+        }
 
         if (encoding) {
             _encoding = encoding;
@@ -341,11 +387,40 @@ function createResponse(options) {
      *  encoding - Optional encoding value.
      */
     mockResponse.end = function(data, encoding) {
+        if (_endCalled) {
+            // Do not emit this event twice.
+            return;
+        }
+
+        mockResponse.finished = true;
+        mockResponse.writableEnded = true;
+        mockResponse.headersSent = true;
 
         _endCalled = true;
 
         if (data) {
-            _data += data;
+            if (data instanceof Buffer) {
+                _chunks.push(data);
+                _size += data.length;
+            } else {
+                _data += data;
+            }
+        }
+
+        if (_chunks.length) {
+            switch (_chunks.length) {
+                case 1:
+                    _buffer = _chunks[0];
+                    break;
+                default:
+                    _buffer = Buffer.alloc(_size);
+                    for (var i = 0, pos = 0, l = _chunks.length; i < l; i++) {
+                        var chunk = _chunks[i];
+                        chunk.copy(_buffer, pos);
+                        pos += chunk.length;
+                    }
+                    break;
+            }
         }
 
         if (encoding) {
@@ -353,6 +428,8 @@ function createResponse(options) {
         }
 
         mockResponse.emit('end');
+        mockResponse.writableFinished = true; //Reference: https://nodejs.org/docs/latest-v12.x/api/http.html#http_request_writablefinished
+        mockResponse.emit('finish');
 
     };
 
@@ -387,23 +464,51 @@ function createResponse(options) {
         return mockResponse.setHeader('Vary', values.join(', '));
     };
 
-   /**
-    * Set header `field` to `val`, or pass
-    * an object of header fields.
-    *
-    * Examples:
-    *
-    *    res.set('Foo', ['bar', 'baz']);
-    *    res.set('Accept', 'application/json');
-    *    res.set({ Accept: 'text/plain', 'X-API-Key': 'tobi' });
-    *
-    * Aliased as `mockResponse.header()`.
-    *
-    * @param {String|Object|Array} field
-    * @param {String} val
-    * @return {ServerResponse} for chaining
-    * @api public
-    */
+    /**
+     * Append additional header `field` with value `val`.
+     *
+     * Example:
+     *
+     *    res.append('Link', ['<http://localhost/>', '<http://localhost:3000/>']);
+     *    res.append('Set-Cookie', 'foo=bar; Path=/; HttpOnly');
+     *    res.append('Warning', '199 Miscellaneous warning');
+     *
+     * @param {String} field
+     * @param {String|Array} val
+     * @return {ServerResponse} for chaining
+     * @api public
+     */
+    mockResponse.append = function append(field, val) {
+        var prev = mockResponse.get(field);
+        var value = val;
+
+        if (prev) {
+            // concat the new and prev vals
+            value = Array.isArray(prev) ? prev.concat(val)
+                : Array.isArray(val) ? [prev].concat(val)
+                    : [prev, val];
+        }
+
+        return mockResponse.set(field, value);
+    };
+
+    /**
+     * Set header `field` to `val`, or pass
+     * an object of header fields.
+     *
+     * Examples:
+     *
+     *    res.set('Foo', ['bar', 'baz']);
+     *    res.set('Accept', 'application/json');
+     *    res.set({ Accept: 'text/plain', 'X-API-Key': 'tobi' });
+     *
+     * Aliased as `mockResponse.header()`.
+     *
+     * @param {String|Object|Array} field
+     * @param {String} val
+     * @return {ServerResponse} for chaining
+     * @api public
+     */
     mockResponse.set = mockResponse.header = function header(field, val) {
         if (arguments.length === 2) {
             if (Array.isArray(val)) {
@@ -412,7 +517,9 @@ function createResponse(options) {
                 val = String(val);
             }
             mockResponse.setHeader(field, val);
-        } else {
+        } else if (typeof field === 'string') {
+            return mockResponse.getHeader(field);
+        } else { // eslint-disable-line
             for (var key in field) {
                 mockResponse.setHeader(key, field[key]);
             }
@@ -420,6 +527,15 @@ function createResponse(options) {
         return mockResponse;
     };
 
+    /**
+     * Function: getHeaders
+     *
+     *   Returns a shallow copy of the current outgoing headers.
+     */
+    mockResponse.getHeaders = function() {
+        return JSON.parse(JSON.stringify(mockResponse._headers));
+    };
+
     /**
      * Function: getHeader
      * Function: get
@@ -427,7 +543,25 @@ function createResponse(options) {
      *   Returns a particular header by name.
      */
     mockResponse.get = mockResponse.getHeader = function(name) {
-        return mockResponse._headers[name];
+        return mockResponse._headers[name.toLowerCase()];
+    };
+
+    /**
+     * Function: getHeaderNames
+     *
+     *   Returns an array containing the unique names of the current outgoing headers.
+     */
+    mockResponse.getHeaderNames = function() {
+        return Object.keys(mockResponse._headers); // names are already stored in lowercase
+    };
+
+    /**
+     * Function: hasHeader
+     *
+     *   Returns `true` if the header identified by `name` is currently set.
+     */
+    mockResponse.hasHeader = function(name) {
+        return name.toLowerCase() in mockResponse._headers;
     };
 
     /**
@@ -437,7 +571,7 @@ function createResponse(options) {
      *   Set a particular header by name.
      */
     mockResponse.setHeader = function(name, value) {
-        mockResponse._headers[name] = value;
+        mockResponse._headers[name.toLowerCase()] = value;
         return value;
     };
 
@@ -447,7 +581,7 @@ function createResponse(options) {
      *   Removes an HTTP header by name.
      */
     mockResponse.removeHeader = function(name) {
-        delete mockResponse._headers[name];
+        delete mockResponse._headers[name.toLowerCase()];
     };
 
     /**
@@ -463,36 +597,33 @@ function createResponse(options) {
         _encoding = encoding;
     };
 
+    mockResponse.getEncoding = function() {
+        return _encoding;
+    };
+
     /**
      * Function: redirect
      *
      *     Redirect to a url with response code
      */
     mockResponse.redirect = function(a, b) {
-
         switch (arguments.length) {
-
             case 1:
-
                 mockResponse.statusCode = 302;
                 _redirectUrl = a;
                 break;
 
             case 2:
-
                 if (typeof a === 'number') {
                     mockResponse.statusCode = a;
                     _redirectUrl = b;
                 }
-
                 break;
 
             default:
                 break;
-
         }
-
-        mockResponse.emit('end');
+        mockResponse.end();
     };
 
     /**
@@ -501,24 +632,70 @@ function createResponse(options) {
      *     Render a view with a callback responding with the
      *     rendered string.
      */
-    mockResponse.render = function(a, b) {
-
+    mockResponse.render = function(a, b, c) {
         _renderView = a;
 
-        switch (arguments.length) {
+        var data = b;
+        var done = c;
 
+        // support callback function as second arg
+        if (typeof b === 'function') {
+            done = b;
+            data = {};
+        }
+
+        switch (arguments.length) {
             case 2:
-                _renderData = b;
+            case 3:
+                _renderData = data;
                 break;
 
             default:
                 break;
-
         }
 
-        mockResponse.emit('render');
-        mockResponse.emit('end');
+        if (typeof done === 'function') {
+            done(null, '');
+        } else {
+            mockResponse.emit('render');
+            mockResponse.end();
+        }
+    };
 
+    /**
+     * Chooses the correct response function from the given supported types.
+     * This method requires that the mockResponse object be initialized with a
+     * mockRequest object reference, otherwise it will throw. @see createMocks.
+     *
+     * @param  {Object} supported Object with formats to handler functions.
+     * @return {Object}           Whatever handler function returns.
+     */
+    mockResponse.format = function(supported) {
+        supported = supported || {};
+        var types = Object.keys(supported);
+
+        if (types.length === 0) {
+            return mockResponse.sendStatus(406);
+        }
+
+        if (!mockRequest) {
+            throw new Error(
+                'Request object unavailable. Use createMocks or pass in a ' +
+                    'request object in createResponse to use format.'
+            );
+        }
+
+        var accepted = mockRequest.accepts(types);
+
+        if (accepted) {
+            return supported[accepted]();
+        }
+
+        if (supported.default) {
+            return supported.default();
+        }
+
+        return mockResponse.sendStatus(406);
     };
 
     // WritableStream.writable is not a function
@@ -538,38 +715,6 @@ function createResponse(options) {
         return writableStream.destroySoon.apply(this, arguments);
     };
 
-    mockResponse.addListener = function() {
-        return eventEmitter.addListener.apply(this, arguments);
-    };
-
-    mockResponse.on = function() {
-        return eventEmitter.on.apply(this, arguments);
-    };
-
-    mockResponse.once = function() {
-        return eventEmitter.once.apply(this, arguments);
-    };
-
-    mockResponse.removeListener = function() {
-        return eventEmitter.removeListener.apply(this, arguments);
-    };
-
-    mockResponse.removeAllListeners = function() {
-        return eventEmitter.removeAllListeners.apply(this, arguments);
-    };
-
-    mockResponse.setMaxListeners = function() {
-        return eventEmitter.setMaxListeners.apply(this, arguments);
-    };
-
-    mockResponse.listeners = function() {
-        return eventEmitter.listeners.apply(this, arguments);
-    };
-
-    mockResponse.emit = function() {
-        return eventEmitter.emit.apply(this, arguments);
-    };
-
     //This mock object stores some state as well
     //as some test-analysis functions:
 
@@ -593,6 +738,15 @@ function createResponse(options) {
         return mockResponse._headers;
     };
 
+    /**
+     *  Function: _getLocals
+     *
+     *  Returns all the locals that were set.
+     */
+    mockResponse._getLocals = function() {
+        return mockResponse.locals;
+    };
+
     /**
      * Function: _getData
      *
@@ -602,6 +756,36 @@ function createResponse(options) {
         return _data;
     };
 
+    /**
+     * Function: _getJSONData
+     *
+     *  The data sent to the user as JSON.
+     */
+    mockResponse._getJSONData = function() {
+        return JSON.parse(_data);
+    };
+
+    /**
+     * Function: _getBuffer
+     *
+     *  The buffer containing data to be sent to the user.
+     *  Non-empty if Buffers were given in calls to write() and end()
+     */
+    mockResponse._getBuffer = function() {
+        return _buffer;
+    };
+
+
+    /**
+     * Function: _getChunks
+     *
+     *  The buffer containing data to be sent to the user.
+     *  Non-empty if Buffers were given in calls to write() and end()
+     */
+    mockResponse._getChunks = function() {
+        return _chunks;
+    };
+
     /**
      * Function: _getStatusCode
      *
@@ -611,6 +795,15 @@ function createResponse(options) {
         return mockResponse.statusCode;
     };
 
+    /**
+     * Function: _getStatusMessage
+     *
+     *  The status message that was sent to the user.
+     */
+    mockResponse._getStatusMessage = function() {
+        return mockResponse.statusMessage;
+    };
+
     /**
      * Function: _isJSON
      *
@@ -618,7 +811,7 @@ function createResponse(options) {
      *  It doesn't validate the data that was sent.
      */
     mockResponse._isJSON = function() {
-        return (mockResponse._headers['Content-Type'] === 'application/json');
+        return mockResponse.getHeader('Content-Type') === 'application/json';
     };
 
     /**
@@ -632,13 +825,10 @@ function createResponse(options) {
      *   False if the encoding wasn't set and wasn't set to "utf8".
      */
     mockResponse._isUTF8 = function() {
-
         if (!_encoding) {
             return false;
         }
-
-        return (_encoding === 'utf8');
-
+        return _encoding === 'utf8';
     };
 
     /**
@@ -654,13 +844,10 @@ function createResponse(options) {
      *   set. Otherwise, it compares it.
      */
     mockResponse._isDataLengthValid = function() {
-
-        if (mockResponse._headers['Content-Length']) {
-            return (mockResponse._headers['Content-Length'].toString() === _data.length.toString());
+        if (mockResponse.getHeader('Content-Length')) {
+            return mockResponse.getHeader('Content-Length').toString() === _data.length.toString();
         }
-
         return true;
-
     };
 
     /**
