@@ -4,6 +4,7 @@ var chai = require('chai');
 var expect = chai.expect;
 var url = require('url');
 var querystring = require('querystring');
+var parseRange = require('range-parser');
 
 var mockRequest = require('../../lib/mockRequest');
 
@@ -32,6 +33,10 @@ describe('mockRequest', function() {
         expect(request).to.be.an('object');
       });
 
+      it('should be an EventEmitter', function() {
+        expect(request).to.be.an.instanceof(require('events').EventEmitter);
+      });
+
       it('should expose Express Request object methods', function() {
         expect(request).to.have.property('get');
         expect(request.get).to.be.a('function');
@@ -47,6 +52,7 @@ describe('mockRequest', function() {
         expect(request.method).to.equal('GET');
         expect(request.url).to.equal('');
         expect(request.originalUrl).to.equal(request.url);
+        expect(request.baseUrl).to.equal(request.url);
         expect(request.path).to.equal('');
         expect(request.params).to.deep.equal({});
         expect(request.session).to.not.exist;
@@ -83,6 +89,7 @@ describe('mockRequest', function() {
         request = mockRequest.createRequest(options);
         expect(request.url).to.equal(options.url);
         expect(request.originalUrl).to.equal(options.url);
+        expect(request.baseUrl).to.equal(options.url);
       });
 
       it('should set .url automatically', function() {
@@ -96,6 +103,15 @@ describe('mockRequest', function() {
         expect(request.url).to.equal(expectedUrl);
       });
 
+      it('should set .baseUrl to options.baseUrl', function() {
+        var options = {
+          baseUrl: '/this'
+        };
+
+        request = mockRequest.createRequest(options);
+        expect(request.baseUrl).to.equal(options.baseUrl);
+      });
+
       it('should set .originalUrl to options.originalUrl', function() {
         var options = {
           originalUrl: '/this/is/a/url'
@@ -352,6 +368,138 @@ describe('mockRequest', function() {
 
   });
 
+  describe('.accepts()', function() {
+    var request;
+
+    beforeEach(function() {
+      request = mockRequest.createRequest({ headers: { accept: 'text/html' }});
+    });
+
+    it('returns type if the same as Accept header', function() {
+      expect(request.accepts('text/html')).to.equal('text/html');
+      expect(request.accepts('html')).to.equal('html');
+    });
+
+    it('returns the first matching type of an array of types', function() {
+      expect(request.accepts(['json', 'html'])).to.equal('html');
+    });
+
+    it('returns false when types dont match', function() {
+      expect(request.accepts(['json', 'xml'])).to.equal(false);
+    });
+  });
+
+  describe('.acceptsEncodings()', function() {
+    var request;
+
+    beforeEach(function() {
+      request = mockRequest.createRequest({ headers: { 'Accept-Encoding': 'gzip' }});
+    });
+
+    it('returns type if the same as Accept-Encoding header', function() {
+      expect(request.acceptsEncodings('gzip')).to.equal('gzip');
+    });
+
+    it('returns the first matching type of an array of types', function() {
+      expect(request.acceptsEncodings(['compress', 'gzip'])).to.equal('gzip');
+    });
+
+    it('returns false when types dont match', function() {
+      expect(request.acceptsEncodings(['compress', 'deflate'])).to.equal(false);
+    });
+  });
+
+  describe('.acceptsCharsets()', function() {
+    var request;
+
+    beforeEach(function() {
+      request = mockRequest.createRequest({ headers: { 'Accept-Charset': 'utf-8' }});
+    });
+
+    it('returns type if the same as Accept-Charset header', function() {
+      expect(request.acceptsCharsets('utf-8')).to.equal('utf-8');
+    });
+
+    it('returns the first matching type of an array of types', function() {
+      expect(request.acceptsCharsets(['iso-8859-15', 'utf-8'])).to.equal('utf-8');
+    });
+
+    it('returns false when types dont match', function() {
+      expect(request.acceptsCharsets(['iso-8859-15', 'us-ascii'])).to.equal(false);
+    });
+  });
+
+  describe('.acceptsLanguages()', function() {
+    var request;
+
+    beforeEach(function() {
+      request = mockRequest.createRequest({ headers: { 'Accept-Language': 'en-GB' }});
+    });
+
+    it('returns type if the same as Accept-Language header', function() {
+      expect(request.acceptsLanguages('en-GB')).to.equal('en-GB');
+    });
+
+    it('returns the first matching type of an array of types', function() {
+      expect(request.acceptsLanguages(['de-DE', 'en-GB'])).to.equal('en-GB');
+    });
+
+    it('returns false when types dont match', function() {
+      expect(request.acceptsLanguages(['de-DE', 'fr-FR'])).to.equal(false);
+    });
+  });
+
+  describe('.range()', function() {
+    var request;
+
+    afterEach(function() {
+      request = null;
+    });
+
+    it('returns undefined if there is no Range header', function() {
+      request = mockRequest.createRequest();
+      expect(request.range()).to.be.undefined;
+    });
+    
+    var tests = [
+      {
+        // Unsatisfiable
+        header: 'bytes=90-100',
+        size: 10
+      },
+      {
+        // Malformed
+        header: 'by90-100',
+        size: 10
+      },
+      {
+        // Basic range
+        header: 'bytes=50-55',
+        size: 100
+      },
+      {
+        // Complex, without options
+        header: 'bytes=50-55,0-10,5-10,56-60',
+        size: 100
+      },
+      {
+        // With options
+        header: 'bytes=50-55,0-10,5-10,56-60',
+        size: 100,
+        options: {
+          combine: true
+        }
+      }
+    ];
+
+    tests.forEach(function(t) {
+      it('returns the result of range-parser if there is a Range header of ' + t.header + ' using size: ' + t.size, function() {
+        request = mockRequest.createRequest({ headers: { range: t.header }});
+        expect(request.range(t.size, t.options)).to.deep.equal(parseRange(t.size, t.header, t.options));
+      });
+    });
+  });
+
   describe('.param()', function() {
     var request;
 
@@ -442,7 +590,6 @@ describe('mockRequest', function() {
       expect(request.get('key')).to.not.exist;
       expect(request.param('key')).to.be.undefined;
     });
-
   });
 
   describe('helper functions', function() {
@@ -596,6 +743,21 @@ describe('mockRequest', function() {
 
     });
 
+    describe('._setBaseUrl()', function() {
+
+      it('should set baseUrl, when called with value', function() {
+        var value = '/path';
+        request._setBaseUrl(value);
+        expect(request.baseUrl).to.equal(value);
+      });
+
+      it('should unset baseUrl, when called with no arguments', function() {
+        request._setBaseUrl();
+        expect(request.baseUrl).to.not.exist;
+      });
+
+    });
+
     describe('._setOriginalUrl()', function() {
 
       it('should set originalUrl, when called with value', function() {
@@ -648,6 +810,108 @@ describe('mockRequest', function() {
 
     });
 
+    describe('.send()', function() {
+
+      it('should trigger data and end event when string is given', function(done) {
+        var data = [];
+        request.on('data', function(chunk) {
+          data.push(chunk);
+        });
+        request.on('end', function() {
+          data = Buffer.concat(data).toString();
+          expect(data).to.equal('test data');
+          done();
+        });
+        request.send('test data');
+      });
+
+      it('should trigger data and end event when object is given', function(done) {
+        var data = [];
+        var dataTosend = {'key' : 'value'};
+        request.on('data', function(chunk) {
+          data.push(chunk);
+        });
+        request.on('end', function() {
+          data = Buffer.concat(data).toString();
+          expect(JSON.parse(data)).to.deep.equal(dataTosend);
+          done();
+        });
+        request.send(dataTosend);
+      });
+
+      it('should trigger data and end event when number is given', function(done) {
+        var data = [];
+        request.on('data', function(chunk) {
+          data.push(chunk);
+        });
+        request.on('end', function() {
+          data = Buffer.concat(data).toString();
+          expect(data).to.equal('35');
+          done();
+        });
+        request.send(35);
+      });
+
+      it('should trigger data and end event when buffer is given', function(done) {
+        var data = [];
+        var bufferdata = Buffer.from('test data');
+        request.on('data', function(chunk) {
+          data.push(chunk);
+        });
+        request.on('end', function() {
+          data = Buffer.concat(data).toString();
+          expect(data).to.equal('test data');
+          done();
+        });
+        request.send(bufferdata);
+      });
+
+      it('should trigger data and end event when nothing is given', function(done) {
+        var data = [];
+        request.on('data', function(chunk) {
+          data.push(chunk);
+        });
+        request.on('end', function() {
+          data = Buffer.concat(data).toString();
+          expect(data).to.equal('');
+          done();
+        });
+        request.send();
+      });
+
+    });
+
+    describe('.subdomains', function() {
+
+      it('should returns the host subdomains', function() {
+        var options = {
+          headers: {
+            host: 'tobi.ferrets.example.com'
+          }
+        };
+        request = mockRequest.createRequest(options);
+        expect(request.subdomains).to.be.an('array').that.includes('ferrets');
+        expect(request.subdomains).to.be.an('array').that.includes('tobi');
+
+      });
+
+      it('should returns and empty array', function () {
+        var options = {
+          headers: {
+            key1: 'key1'
+          }
+        };
+        request = mockRequest.createRequest(options);
+        expect(request.subdomains).to.be.an('array').that.empty;
+
+        options.headers.host = 'example.com';
+        request = mockRequest.createRequest(options);
+        expect(request.subdomains).to.be.an('array').that.empty;
+
+      });
+
+    });
+
   });
 
 });
